from collections.abc import Sequence
from typing import Any, Union

from pydantic import TypeAdapter

from ..driver import Driver
from ..state import DB
from .base import Package, PackageArgs
from .brew_repo import BrewRepo
from .clone import Clone
from .daemon import Daemon
from .flatpak import FlatpakPackage
from .links import Links
from .npm import NpmPackage
from .shell import Shell
from .system import SystemPackage
from .yum_repo import YumRepo

AnyPackageT = Union[
    BrewRepo,
    Clone,
    Daemon,
    FlatpakPackage,
    Links,
    NpmPackage,
    Shell,
    SystemPackage,
    YumRepo,
]

AnyPackage: TypeAdapter[AnyPackageT] = TypeAdapter(AnyPackageT)

AnyPackageList = TypeAdapter(list[AnyPackageT])


def parse_package(package: Any, *, db: DB, driver: Driver) -> Package:
    package = TypeAdapter(dict).validate_python(package)

    return AnyPackage.validate_python({**package, "db": db, "driver": driver})


def parse_packages(packages: Any, *, db: DB, driver: Driver) -> Sequence[Package]:
    packages = TypeAdapter(list[dict]).validate_python(packages)

    return AnyPackageList.validate_python(
        [{**package, "db": db, "driver": driver} for package in packages]
    )



from typing import Optional

from .base import PackageCacheInstaller, PackageVersionInfo


class DNF(PackageCacheInstaller):
    async def install(self, package: str) -> None:
        await self.driver.with_root(True).run("dnf", "install", "-y", package)
        await super().install(package)

    # can't install an old version of a package in tests
    async def upgrade(self, package: str) -> None:  # pragma: no cover
        await self.driver.with_root(True).run("dnf", "upgrade", "-y", package)
        await super().upgrade(package)

    @staticmethod
    def parse_versions(output: str, package: Optional[str]) -> dict[str, str]:
        versions = {}
        for line in output.splitlines():
            name, version = line.split()
            versions[name] = version

        if package is not None:
            if package in versions:
                return {package: versions[package]}
            # If querying for a specific package, the name might be different
            # because of virtual packages and --whatprovides option.
            # If there's no package with exact name, require a single match.
            if len(versions) > 1:
                raise ValueError(
                    f"Multiple versions for {package}: {versions}."
                )  # pragma: no cover
            if len(versions) == 0:
                raise ValueError(f"No versions for {package}.")
            (version,) = versions.values()  # pylint:disable=unbalanced-dict-unpacking
            return {package: version}

        return versions

    async def rpm_query(self, package: Optional[str]) -> dict[str, str]:
        """
        Query RPM for the installed package versions.

        @param package: If specified, only query this package; otherwise,
        return all packages' versions.
        """
        args = [
            "rpm",
            "--query",
            "--queryformat",
            "%{NAME} %{VERSION}\n",
        ]

        if package:
            args += ["--whatprovides", package]
        else:
            args += ["--all"]

        result = await self.driver.run_output_(*args)
        if not result.ok:
            return {}
        return self.parse_versions(result.output, package)

    async def dnf_repoquery(self, package: Optional[str]) -> dict[str, str]:
        """
        Query DNF for the latest package versions.

        @param package: If specified, only query this package; otherwise,
        return all packages' versions.
        """

        await self.driver.run("dnf", "--quiet", "clean", "expire-cache")

        arch = await self.driver.architecture()

        args = [
            "dnf",
            "--quiet",
            "repoquery",
            "--queryformat",
            "%{name} %{version}\n",
            "--latest-limit",
            "1",
            "--arch",
            f"{arch},noarch",
        ]

        if package:
            args += ["--whatprovides", package]

        output = await self.driver.run_output(*args)
        return self.parse_versions(output, package)

    async def get_package_info(
        self, package: Optional[str]
    ) -> dict[str, PackageVersionInfo]:
        latest = await self.dnf_repoquery(package)
        installed = await self.rpm_query(package)

        return {
            name: PackageVersionInfo(
                installed=installed.get(name), latest=latest_version
            )
            for name, latest_version in latest.items()
        }



import sqlite3
import threading
from pathlib import Path
from types import TracebackType

DB_PATH = "state.sqlite"


class DB:
    """Thread-safe SQLite database connection."""

    @classmethod
    def temporary(cls) -> "DB":
        return cls(":memory:", close=False)

    def __init__(self, path: str | Path, close: bool = True) -> None:
        self.path = path
        self.cache = threading.local()
        self.close = close

    def _make_connection(self) -> sqlite3.Connection:
        connection = sqlite3.connect(self.path)
        connection.row_factory = sqlite3.Row
        return connection

    @property
    def _connection(self) -> sqlite3.Connection:
        if not hasattr(self.cache, "connection"):
            self.cache.connection = self._make_connection()
        return self.cache.connection

    @property
    def _depth(self) -> int:
        return getattr(self.cache, "depth", 0)

    def _depth_alter(self, delta: int) -> int:
        self.cache.depth = self._depth + delta
        return self.cache.depth

    def __enter__(self) -> sqlite3.Connection:
        self._depth_alter(1)
        return self._connection.__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._connection.__exit__(exc_type, exc_val, exc_tb)
        if self._depth_alter(-1) == 0:
            if self.close:
                self._connection.close()
                delattr(self.cache, "connection")
